#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <utime.h>
#include <dlfcn.h>
#include <errno.h>

void *handle = NULL;

#define MAXLENGTH 1024

static char *home_etc_pathname(const char *pathname)
{
	char buf[MAXLENGTH];
	char *buf2, *end, *home, *home_etc;
	int len;

/*fprintf(stderr, "home_etc_pathname: pathname = %s\n", pathname);*/
	if (!pathname) return NULL;
	if (pathname[0] == '/') {
		end = strstr(pathname, "/.");
		if (!end) return NULL;
		strncpy(buf, pathname, MAXLENGTH);
		end = buf + (end - pathname);
	} else {
		getcwd(buf, MAXLENGTH);
		len = strlen(buf);
		strncpy(buf + len, pathname, MAXLENGTH - len);
		end = strstr(buf, "/.");
		if (!end) return NULL;
	}
	
	home = getenv("HOME");
	if (!home) return NULL;
	home_etc = getenv("HOME_ETC");
	if (!home_etc) return NULL;
	*end = '\0';
	if (strcmp(buf, home)) return NULL;
	buf2 = (unsigned char *)malloc(MAXLENGTH);
	if (!buf2) return NULL;
	snprintf(buf2, MAXLENGTH, "%s/%s", home_etc, end + 1);
	/*fprintf(stderr, "buf2 = %s\n", buf2);*/
	return buf2;	
}
#undef MAXLENGTH

int access(const char *pathname, int how)
{
	static int (*func)(const char *, int) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "access");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(pathname);
	if (!new_pathname) return (*func)(pathname, how);
	else {
		int res = (*func)(new_pathname, how);

		free(new_pathname);
		return res;
	}
}

int bind(int fd, __CONST_SOCKADDR_ARG addr, socklen_t addrlen)
{
	static int (*func)(int, __CONST_SOCKADDR_ARG, socklen_t) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "bind");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	if (!addr) {
		errno = ENOMEM;
		return -1;
	}
	if (addr->sa_family == AF_UNIX) {
		struct sockaddr_un new_addr, *addr2 = (struct sockaddr_un *)addr;
		int res, len;
		
		new_pathname = home_etc_pathname(addr2->sun_path);
		if (!new_pathname) return (*func)(fd, addr, addrlen);
		len = strlen(new_pathname);
		if (len > 107) {
			errno = ENOMEM;
			free(new_pathname);
			return -1;
		}

		addrlen = sizeof(new_addr.sun_family) + len;
		bzero((char *)&new_addr, sizeof(new_addr));
		new_addr.sun_family = AF_UNIX;
		strcpy(new_addr.sun_path, new_pathname);

		res = (*func)(fd, (__CONST_SOCKADDR_ARG)&new_addr, addrlen);
		free(new_pathname);
		return res;
	} else {
		return (*func)(fd, addr, addrlen);
	}
}

int chdir(const char *pathname)
{
	static int (*func)(const char *) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "chdir");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(pathname);
	if (!new_pathname) return (*func)(pathname);
	else {
		int res = (*func)(new_pathname);

		free(new_pathname);
		return res;
	}
}

int chmod(const char *pathname, mode_t mode)
{
	static int (*func)(const char *, mode_t) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "chmod");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(pathname);
	if (!new_pathname) return (*func)(pathname, mode);
	else {
		int res = (*func)(new_pathname, mode);

		free(new_pathname);
		return res;
	}
}

int chown(const char *pathname, uid_t owner, gid_t group)
{
	static int (*func)(const char *, uid_t, gid_t) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "chown");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(pathname);
	if (!new_pathname) return (*func)(pathname, owner, group);
	else {
		int res = (*func)(new_pathname, owner, group);

		free(new_pathname);
		return res;
	}
}

int connect(int fd, const struct sockaddr *addr, socklen_t addrlen)
{
	static int (*func)(int, const struct sockaddr *, socklen_t) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "connect");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	if (!addr) {
		errno = ENOMEM;
		return -1;
	}
	if (addr->sa_family == AF_UNIX) {
		struct sockaddr_un new_addr, *addr2 = (struct sockaddr_un *)addr;
		int res, len;
		
		new_pathname = home_etc_pathname(addr2->sun_path);
		if (!new_pathname) return (*func)(fd, addr, addrlen);
		len = strlen(new_pathname);
		if (len > 107) {
			errno = ENOMEM;
			free(new_pathname);
			return -1;
		}

		addrlen = sizeof(new_addr.sun_family) + len;
		bzero((char *)&new_addr, sizeof(new_addr));
		new_addr.sun_family = AF_UNIX;
		strcpy(new_addr.sun_path, new_pathname);

		res = (*func)(fd, (const struct sockaddr *)&new_addr, addrlen);
		free(new_pathname);
		return res;
	} else {
		return (*func)(fd, addr, addrlen);
	}
}

int creat(const char *pathname, mode_t mode)
{
	static int (*func)(const char *, mode_t) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "creat");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(pathname);
	if (!new_pathname) return (*func)(pathname, mode);
	else {
		int res = (*func)(new_pathname, mode);

		free(new_pathname);
		return res;
	}
}

int creat64(const char *pathname, mode_t mode)
{
	static int (*func)(const char *, mode_t) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "creat64");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(pathname);
	if (!new_pathname) return (*func)(pathname, mode);
	else {
		int res = (*func)(new_pathname, mode);

		free(new_pathname);
		return res;
	}
}

FILE* fopen(const char *filename, const char *mode)
{
	static FILE *(*func)(const char *, const char *) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return NULL;
	}
	if (!func) {
		func = dlsym(handle, "fopen");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return NULL;
		}
	}
	new_pathname = home_etc_pathname(filename);
	
	if (!new_pathname) return (*func)(filename, mode);
	else {
		FILE *res = (*func)(new_pathname, mode);

		free(new_pathname);
		return res;
	}
}

FILE* freopen(const char *filename, const char *mode, FILE *stream)
{
	static FILE *(*func)(const char *, const char *, FILE *) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return NULL;
	}
	if (!func) {
		func = dlsym(handle, "freopen");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return NULL;
		}
	}
	new_pathname = home_etc_pathname(filename);
	
	if (!new_pathname) return (*func)(filename, mode, stream);
	else {
		FILE *res = (*func)(new_pathname, mode, stream);

		free(new_pathname);
		return res;
	}
}

int lchmod(const char *pathname, mode_t mode)
{
	static int (*func)(const char *, mode_t) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "lchmod");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(pathname);
	if (!new_pathname) return (*func)(pathname, mode);
	else {
		int res = (*func)(new_pathname, mode);

		free(new_pathname);
		return res;
	}
}

int lchown(const char *pathname, uid_t owner, gid_t group)
{
	static int (*func)(const char *, uid_t, gid_t) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "lchown");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(pathname);
	if (!new_pathname) return (*func)(pathname, owner, group);
	else {
		int res = (*func)(new_pathname, owner, group);

		free(new_pathname);
		return res;
	}
}

int lstat(const char *filename, struct stat *buf)
{
	static int (*func)(const char *, struct stat *) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "lstat");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(filename);
	if (!new_pathname) return (*func)(filename, buf);
	else {
		int res = (*func)(new_pathname, buf);

		free(new_pathname);
		return res;
	}
}

int lstat64(const char *filename, struct stat64 *buf)
{
	static int (*func)(const char *, struct stat64 *) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "lstat64");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(filename);
	if (!new_pathname) return (*func)(filename, buf);
	else {
		int res = (*func)(new_pathname, buf);

		free(new_pathname);
		return res;
	}
}

int mkdir(const char *pathname, mode_t mode)
{
	static int (*func)(const char *, mode_t) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "mkdir");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(pathname);
	if (!new_pathname) return (*func)(pathname, mode);
	else {
		int res = (*func)(new_pathname, mode);

		free(new_pathname);
		return res;
	}
}

int mkfifo(const char *pathname, mode_t mode)
{
	static int (*func)(const char *, mode_t) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "mkfifo");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(pathname);
	if (!new_pathname) return (*func)(pathname, mode);
	else {
		int res = (*func)(new_pathname, mode);

		free(new_pathname);
		return res;
	}
}

int mknod(const char *pathname, mode_t mode, dev_t dev)
{
	static int (*func)(const char *, mode_t, dev_t) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "mknod");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(pathname);
	if (!new_pathname) return (*func)(pathname, mode, dev);
	else {
		int res = (*func)(new_pathname, mode, dev);

		free(new_pathname);
		return res;
	}
}

int open(const char *filename, int flags, ...)
{
	static int (*func)(const char *, int, ...) = NULL;
	char *error;
	unsigned char *new_pathname;
	va_list ap;
	mode_t mode = -1;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "open");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(filename);

	if (flags & O_CREAT) {
		va_start(ap, flags);
		mode = va_arg(ap, mode_t);
		va_end(ap);
	}
	
	if (!new_pathname) return (*func)(filename, flags, mode);
	else {
		int res = (*func)(new_pathname, flags, mode);

		free(new_pathname);
		return res;
	}
}

int open64(const char *filename, int flags, ...)
{
	static int (*func)(const char *, int, ...) = NULL;
	char *error;
	unsigned char *new_pathname;
	va_list ap;
	mode_t mode = 0;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "open64");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(filename);

	if (flags & O_CREAT) {
		va_start(ap, flags);
		mode = va_arg(ap, mode_t);
		va_end(ap);
	}

	if (!new_pathname) return (*func)(filename, flags, mode);
	else {
		int res = (*func)(new_pathname, flags, mode);

		free(new_pathname);
		return res;
	}
}

int stat(const char *filename, struct stat *buf)
{
	static int (*func)(const char *, struct stat *) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "stat");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(filename);
	if (!new_pathname) return (*func)(filename, buf);
	else {
		int res = (*func)(new_pathname, buf);

		free(new_pathname);
		return res;
	}
}

int stat64(const char *filename, struct stat64 *buf)
{
	static int (*func)(const char *, struct stat64 *) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "stat64");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(filename);
	if (!new_pathname) return (*func)(filename, buf);
	else {
		int res = (*func)(new_pathname, buf);

		free(new_pathname);
		return res;
	}
}

int utime(const char *filename, const struct utimbuf *buf)
{
	static int (*func)(const char *, const struct utimbuf *) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "utime");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(filename);
	if (!new_pathname) return (*func)(filename, buf);
	else {
		int res = (*func)(new_pathname, buf);

		free(new_pathname);
		return res;
	}
}

int utimes(const char *filename, struct timeval *tvp)
{
	static int (*func)(const char *, struct timeval *) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "utimes");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(filename);
	if (!new_pathname) return (*func)(filename, tvp);
	else {
		int res = (*func)(new_pathname, tvp);

		free(new_pathname);
		return res;
	}
}

int _xstat(int ver, const char *filename, struct stat *buf)
{
	static int (*func)(int, const char *, struct stat *) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "_xstat");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(filename);
	if (!new_pathname) return (*func)(ver, filename, buf);
	else {
		int res = (*func)(ver, new_pathname, buf);

		free(new_pathname);
		return res;
	}
}

int _xstat64(int ver, const char *filename, struct stat64 *buf)
{
	static int (*func)(int, const char *, struct stat64 *) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "_xstat64");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(filename);
	if (!new_pathname) return (*func)(ver, filename, buf);
	else {
		int res = (*func)(ver, new_pathname, buf);

		free(new_pathname);
		return res;
	}
}

int __xstat(int ver, const char *filename, struct stat *buf)
{
	static int (*func)(int, const char *, struct stat *) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "__xstat");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(filename);
	if (!new_pathname) return (*func)(ver, filename, buf);
	else {
		int res = (*func)(ver, new_pathname, buf);

		free(new_pathname);
		return res;
	}
}

int __xstat64(int ver, const char *filename, struct stat64 *buf)
{
	static int (*func)(int, const char *, struct stat64 *) = NULL;
	char *error;
	unsigned char *new_pathname;

	if (!handle) handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
/* ENOMEM is fake, but something should be set */
	if (!handle) {
		errno = ENOMEM;
		return -1;
	}
	if (!func) {
		func = dlsym(handle, "__xstat64");
		if ((error = dlerror()) != NULL) {
			errno = ENOMEM;
			return -1;
		}
	}
	new_pathname = home_etc_pathname(filename);
	if (!new_pathname) return (*func)(ver, filename, buf);
	else {
		int res = (*func)(ver, new_pathname, buf);

		free(new_pathname);
		return res;
	}
}
